using System.Collections;
using SpellBoundAR.QRCodes;
using Unity.Collections;
using UnityEngine;
using UnityEngine.XR.ARFoundation;
using UnityEngine.XR.ARSubsystems;
using ZXing;

namespace SpellBoundAR.ARImageTracking
{
    public class ARFoundationQRCodeScanner : QRCodeScanner
    {
        [Header("References")]
        [SerializeField] private ARCameraManager arCameraManager;

        [Header("Cache")]
        private bool _isProcessing;
        private XRCpuImage _currentCPUImage;
        
        private bool _cachedDimensions;
        private float _cachedWidth;
        private float _cachedHeight;

        public override float Width => _cachedDimensions ? _cachedWidth : Screen.width;
        public override float Height => _cachedDimensions ? _cachedHeight : Screen.height;

        public override Result Scan()
        {
            if (_isProcessing) return null;
            if (!arCameraManager.TryAcquireLatestCpuImage(out _currentCPUImage)) return null;
            
            _isProcessing = true;
            _cachedDimensions = true;
            _cachedWidth = _currentCPUImage.width;
            _cachedHeight = _currentCPUImage.height;
            StartCoroutine(ProcessImage(_currentCPUImage));
            _currentCPUImage.Dispose();

            return null;
        }

        IEnumerator ProcessImage(XRCpuImage image)
        {
            Debug.LogWarning("ProcessQRCode image.dimensions: " + image.dimensions);

            // Create the async conversion request.
            var request = image.ConvertAsync(new XRCpuImage.ConversionParams
            {
                // Use the full image.
                inputRect = new RectInt(0, 0, image.width, image.height),

                // Downsample by 2.
                //outputDimensions = new Vector2Int(image.width / 2, image.height / 2),
                outputDimensions = new Vector2Int(image.width, image.height),

                // Color image format.
                outputFormat = TextureFormat.R8,

                // Flip across the Y axis.
                transformation = XRCpuImage.Transformation.MirrorY
            });

            // Wait for the conversion to complete.
            while (!request.status.IsDone())
                yield return null;

            // Check status to see if the conversion completed successfully.
            if (request.status != XRCpuImage.AsyncConversionStatus.Ready)
            {
                // Something went wrong.
                Debug.LogWarningFormat("ProcessImage: Request failed with status {0}", request.status);

                // Dispose even if there is an error.
                request.Dispose();
                yield break;
            }

            // Image data is ready. Let's apply it to a Texture2D.
            var rawData = request.GetData<byte>();


            if (ShowDebugTexture)
            {
                // Create a texture if necessary.
                if (!DebugTexture)
                {
                    DebugTexture = new Texture2D(
                        request.conversionParams.outputDimensions.x,
                        request.conversionParams.outputDimensions.y,
                        request.conversionParams.outputFormat,
                        false);
                }

                // Copy the image data into the texture.
                DebugTexture.LoadRawTextureData(rawData);
                DebugTexture.Apply();
            }


            byte[] grayscalePixels = NativeArrayExtensions.ToRawBytes(rawData);
            
            // The 'grayscalePixels' array now contains the image data.
            var result = BarcodeReader.Decode(grayscalePixels, image.width, image.height, RGBLuminanceSource.BitmapFormat.Gray8);
            
            if (result != null)
            {
                InvokeOnQRCodeScanned(result);
            }

            // Need to dispose the request to delete resources associated
            // with the request, including the raw data.
            request.Dispose();


            _isProcessing = false;
        }
    }

    /// <summary>
    /// https://gist.github.com/asus4/992ae563ca1263b1dd936b970e7fc206
    /// </summary>
    public static class NativeArrayExtensions
    {
        public static byte[] ToRawBytes<T>(this NativeArray<T> arr) where T : struct
        {
            var slice = new NativeSlice<T>(arr).SliceConvert<byte>();
            var bytes = new byte[slice.Length];
            slice.CopyTo(bytes);
            return bytes;
        }
    }
}
